using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Text;
using NUnit.Framework;

namespace Concurrency
{
    internal class TotalsInLoopReader : IStressTestable
    {
        private SqlCommand _command;
        private readonly int[] numbers = new int[2000];
        private int currentIndex = 0;
        private int errors = 0;

        public void Prepare(SqlConnection connection)
        {
            _command = connection.CreateCommand();
            _command.CommandText = "Readers.SelectTotalTickets";
            _command.CommandType = CommandType.StoredProcedure;
        }

        public void Run()
        {
            DateTime startedAt = DateTime.Now;
            for (int i = 0; i < numbers.Length; i++)
            {
                try
                {
                    //the commented line below is incorrect:
                    //if _command.ExecuteScalar() raises an exception, currentIndex is already incremented,
                    // so there is a gap in numbers array
                    //numbers[currentIndex++] = Convert.ToInt32(_command.ExecuteScalar());
                    int total = Convert.ToInt32(_command.ExecuteScalar());
                    numbers[currentIndex++] = total;
                }
                catch (Exception e)
                {
                    Console.WriteLine(e);
                    errors++;
                }
                if (errors > 10)
                {
                    break;
                }
            }
            Console.WriteLine(
                string.Format("TotalsInLoopReader: {0} completed, {1} failed, duration {2}", currentIndex, errors,
                              DateTime.Now.Subtract(startedAt)));
            int incorrectTotalsCounter = 0;
            for (int i = 0; i < currentIndex; i++)
            {
                if (numbers[i] != 10000)
                {
                    Console.WriteLine(string.Format("TotalsInLoopReader Iteration {0}, incorrect totals :{1}", i, numbers[i]));
                    incorrectTotalsCounter++;
                }
            }
            Console.WriteLine(
                string.Format("TotalsInLoopReader: {0} incorrect totals out of {1}", incorrectTotalsCounter,
                              currentIndex
                    ));
        }
    }

    internal class TicketsShuffler : IStressTestable
    {
        private SqlCommand _command;
        private int errors = 0;

        public void Prepare(SqlConnection connection)
        {
            _command = connection.CreateCommand();
            _command.CommandText = "Writers.ReassignTicket";
            _command.CommandType = CommandType.StoredProcedure;
            _command.Parameters.AddWithValue("@TicketId", 0);
            _command.Parameters.AddWithValue("@AssignedTo", 0);
        }

        public void Run()
        {
            DateTime startedAt = DateTime.Now;
            Random random = new Random();
            const int numIterations = 8000;
            for (int i = 0; i < numIterations; i++)
            {
                try
                {
                    _command.Parameters[0].Value = Convert.ToInt32(random.Next(1, 10000));
                    _command.Parameters[1].Value = Convert.ToInt32(random.Next(1, 10000));
                    _command.ExecuteNonQuery();
                }
                catch (Exception e)
                {
                    Console.WriteLine(e);
                    errors++;
                }
                if (errors > 10)
                {
                    break;
                }
            }
            Console.WriteLine(
                string.Format("TicketsShuffler: {0} completed, {1} failed, duration {2}", numIterations - errors, errors,
                              DateTime.Now.Subtract(startedAt)));
        }
    }

    [TestFixture]
    public class TotalsTests
    {
        readonly string connectionString = "server=(local);trusted_connection=true;database=StressTests";

        [Test]
        public void TotalsAndShufflerTest()
        {
            using (SqlConnection connection = new SqlConnection(connectionString))
            {
                connection.Open();
                SqlCommand command = connection.CreateCommand();
                command.CommandText = "Writers.RepopulateTickets";
                command.CommandType = CommandType.StoredProcedure;
                command.CommandTimeout = 0;
                command.ExecuteNonQuery();
                using (
                    StressTester tester =
                        new StressTester(connectionString, new TotalsInLoopReader(), new TicketsShuffler()))
                {
                    tester.RunConcurrently();
                }
            }
        }
    }
}
